#####################################################
###   A registry of callable view functions
###   in the DMP system

from django.conf import settings
from django.apps import apps, AppConfig

from .template import render_to_string_shortcut, render_shortcut
from .util import get_dmp_instance

import threading


# lock to keep register_app() thread safe
rlock = threading.RLock()

# the DMP-enabled AppConfigs in the system
DMP_ENABLED_APPS = set()


###############################################################
###   Registration functions for apps and view functions
###

def is_dmp_app(app):
    '''
    Returns True if the given app is a DMP-enabled app.  The app parameter can
    be either the name of the app or an AppConfig object.
    '''
    if isinstance(app, AppConfig):
        app = app.name
    return app in DMP_ENABLED_APPS


def register_app(app):
    '''
    Registers an app as a "DMP-enabled" app.  Registering creates a cached
    template renderer to make processing faster and adds the dmp_render()
    and dmp_render_to_string() methods to the app.   The app parameter can
    be either the name of the app or an AppConfig object.

    This is called by MakoTemplates (engine.py) during system startup.
    '''
    if isinstance(app, str):
        app = apps.get_app_config(app)

    with rlock:
        # short circuit if already in the DMP_ENABLED_APPS
        if app.name in DMP_ENABLED_APPS:
            return

        # check for the DJANGO_MAKO_PLUS flag in the module
        if not getattr(app.module, 'DJANGO_MAKO_PLUS', False):
            return

        # first time for this app, so add to our dictionary
        # defaultdict, so no need to set to anything
        DMP_ENABLED_APPS.add(app.name)

        # set up the template, script, and style renderers
        # these create and cache just by accessing them
        get_dmp_instance().get_template_loader(app, 'templates', create=True)
        get_dmp_instance().get_template_loader(app, 'scripts', create=True)
        get_dmp_instance().get_template_loader(app, 'styles', create=True)

        # add the shortcut functions (only to the main templates, we don't do to scripts or styles
        # because people generally don't call those directly).  This is a monkey patch, but it is
        # an incredibly useful one because it makes calling app-specific rendering functions much
        # easier.
        #
        # Django's shortcut to return an *HttpResponse* is render(), and its template method to render a *string* is also render().
        # Good job on naming there, folks.  That's going to confuse everyone.  But I'm matching it to be consistent despite the potential confusion.
        app.module.dmp_render_to_string = render_to_string_shortcut(app.name)
        app.module.dmp_render = render_shortcut(app.name)





#
#   Author:  Conan Albrecht <ca&byu,edu>
#   License: Apache Open Source License
#


# pointer to our app config
# Django looks for this exact variable name
default_app_config = 'django_mako_plus.Config'


# the version
from .version import __version__


# the app config
from .apps import Config


# the router and middleware
from .middleware import RequestInitMiddleware
from .router import route_request


# view_function decorator and converter classes
from .router import view_function
from .converter import DefaultConverter, set_default_converter, get_default_converter


# the middleware and template
# the template engine
from .engine import MakoTemplates


# the exceptions
from .exceptions import RedirectException
from .exceptions import PermanentRedirectException
from .exceptions import JavascriptRedirectException
from .exceptions import InternalRedirectException


# filters
from .filters import django_syntax, jinja2_syntax, alternate_syntax


# the http responses
from .http import HttpResponseJavascriptRedirect


# the convenience functions
#
# Instead of these functions, consider using dmp_render() and dmp_render_to_string(),
# which are monkey-patched onto every DMP-enabled app at load time.  See the documentation
# for information on why we do this.
#
from .convenience import render_template
from .convenience import render_template_for_path
from .convenience import get_template
from .convenience import get_template_for_path
from .convenience import get_template_loader
from .convenience import get_template_loader_for_path


# the utilities
from .util import get_dmp_instance
from .util import get_dmp_app_configs


# the urls
# I'm specifically not including urls.py here because I want it imported
# as late as possible (after all the apps are set up).  Django will import it
# when it processes the project's urls.py file.


# the static files shortcuts
from .static_files import link_css
from .static_files import link_js
from .static_files import link_template_css
from .static_files import link_template_js
# these are deprecated as of Jan 2017 and can be removed at some point
from .static_files import get_template_css
from .static_files import get_template_js



from django.conf.urls import url
try:
    from django.urls.resolvers import RegexURLPattern     # Django 1.10+
    from django.urls.exceptions import Resolver404
except ImportError:
    from django.core.urlresolvers import RegexURLPattern  # Django 1.9
    from django.core.urlresolvers import Resolver404
from .router import route_request
from .registry import is_dmp_app



#########################################################
###   A custom url pattern that checks for DMP apps.

class DMPRegexPattern(RegexURLPattern):
    def resolve(self, path):
        '''
        First does the regular resolve.  If a match is found on the pattern,
        it then checks whether the named "dmp_router_app" variable is a
        DMP app.  If it is not a DMP app, it raises a Resolver404 indicating
        that we do not have a match.

        I'm doing this here in the url resolution instead of within route_request
        because I need to allow some overlapping patterns like /app and /page.

        An alternative to this would be to create a set of url patterns for each app.
        This would prevent this custom resolver, but it would significantly increase the
        number of url patterns (ugly).  And it would only include DMP apps that exist at
        compile time (and not any that are registered afterward, although I'm not sure
        when/if this would ever happen).

        I'm still not sure I made the right decision in choosing this small hack instead
        of the big set of normal url patterns.  But this is the decision for now.

        Any pattern wrapped in this class MUST have the named parameter: (?P<dmp_router_app>...)
        '''
        # let the super do its work
        rmatch = super().resolve(path)

        # if we have a match, check that the dmp_router_app is DMP-enabled
        if rmatch is not None and not is_dmp_app(rmatch.kwargs['dmp_router_app']):
            raise Resolver404({'path': path})

        # return
        return rmatch




#########################################################
###   The default DMP url patterns

# FYI, even though the valid python identifier is [_A-Za-z][_a-zA-Z0-9]*, I'm simplifying it to [_a-zA-Z0-9]+ because it works for our purposes

urlpatterns = [
    # these are in order of specificity, with the most specific ones at the top

    # /app/page.function/urlparams
    DMPRegexPattern(r'^(?P<dmp_router_app>[_a-zA-Z0-9\-]+)/(?P<dmp_router_page>[_a-zA-Z0-9\-]+)\.(?P<dmp_router_function>[_a-zA-Z0-9\.\-]+)/?(?P<urlparams>.*?)/?$', route_request, name='DMP /app/page.function'),

    # /app/page/urlparams
    DMPRegexPattern(r'^(?P<dmp_router_app>[_a-zA-Z0-9\-]+)/(?P<dmp_router_page>[_a-zA-Z0-9\-]+)/?(?P<urlparams>.*?)/?$', route_request, name='DMP /app/page'),

    # /app
    # FYI: /app/urlparams can't happen because the first urlparam would be captured as /app/page in the previous pattern
    DMPRegexPattern(r'^(?P<dmp_router_app>[_a-zA-Z0-9\-]+?)/?$', route_request, name='DMP /app'),

    # /page.function/urlparams
    url(r'^(?P<dmp_router_page>[_a-zA-Z0-9\-]+)\.(?P<dmp_router_function>[_a-zA-Z0-9\.\-]*)/?(?P<urlparams>.*?)/?$', route_request, name='DMP /page.function'),

    # /page/urlparams
    url(r'^(?P<dmp_router_page>[_a-zA-Z0-9\-]+)/?(?P<urlparams>.*?)/?$', route_request, name='DMP /page'),

    # / with nothing else
    # FYI: /urlparams can't happen because it would be captured as /page in the previous pattern
    url(r'^$', route_request, name='DMP /'),
]




from django.conf import settings
from django.http import Http404
from django.db.models import Model, ObjectDoesNotExist

from .exceptions import RedirectException
from .util import DMP_OPTIONS, log

import inspect, datetime, decimal, sys
from collections import namedtuple
from operator import attrgetter


##################################################################
###   ConversionTask object

class ConversionTask(object):
    '''
    A (mostly) data class that holds meta-information about a conversion
    task.  This object is sent into each converter function.
    '''
    def __init__(self, request, module, function, kwargs):
        self.converter = _check_converter(kwargs.get('converter'))
        self.request = request
        self.module = module
        self.function = function
        self.kwargs = kwargs       # kwargs from the @view_function decorator


##################################################################
###   ConverterInfo class: see @convert_method below

class ConverterInfo(object):
    '''
    Used by @convert_method to track types on a converter method.
    '''
    CONVERT_TYPES_KEY = '_dmp_converter_infos'
    DECLARED_ORDER = 0

    def __init__(self, convert_type, convert_func):
        self.convert_type = convert_type
        self.convert_func = convert_func
        # we sort by reversed( len(mro), DECLARED_ORDER ) so subclasses match first
        self.sort_key = ( -1 * len(inspect.getmro(convert_type)), -1 * ConverterInfo.DECLARED_ORDER )
        ConverterInfo.DECLARED_ORDER = 0 if ConverterInfo.DECLARED_ORDER == sys.maxsize else ConverterInfo.DECLARED_ORDER + 1


##################################################################
###   Default converter class


class BaseConverter(object):

    @staticmethod
    def convert_method(*class_types):
        '''
        Decorator for converter methods in DefaultConverter.
        '''
        def inner(func):
            setattr(func, ConverterInfo.CONVERT_TYPES_KEY, [ ConverterInfo(ct, func) for ct in class_types ])
            return func
        return inner


    def __init__(self):
        # Discover the converters in this object - these were placed here by the @convert_method decorator.
        self.converters = []
        for name, method in inspect.getmembers(self, inspect.ismethod):
            self.converters.extend(getattr(method, ConverterInfo.CONVERT_TYPES_KEY, ()))
        self.converters.sort(key=attrgetter('sort_key'))


    def __call__(self, value, parameter, task):
        '''
        Converts the given value to the type expected by the view function.

            value:      value from request.urlparams to convert
                        The value will always be a string, even if empty '' (never None).

            parameter:  an instance of ViewParameter.  See that class above for more information.

            task:       meta-information about the current conversion task (as a named tuple):
                        module:             A reference to the module containing the view function.
                        function:           A reference to the view function.
                        converter:          A reference to the currently-running converter callable.
                        urlparams:          The unconverted request.urlparams list of strings from the url,
                                            provided for cases when converting a value depends on another.

        This method should return one of the following:

            1. Converted value (desired type is in parameter.type) to be placed in converted_args.

            2. An HttpResponse object, which is immediately returned to the browser (without further processing).

            3. Raise an exception to signal an error or validation problem, such as:
                a. Django's Http404 exception (returns immediately to the browser)
                b. DMP's RedirectException (sends a redirect to the browser)
                c. DMP's InternalRedirectException (internally restarts the router with a new view function)
               See the @view_parameter decorator for more information on handling conversion errors.
        '''
        # we don't convert anything without type hints
        if parameter.type is inspect.Parameter.empty:
            return value

        # if the value is already the right type, return it
        if isinstance(value, parameter.type):
            return value

        # find the converter method for this type
        # I'm iterating through the list to find the most specific match first
        # The list is sorted by specificity so subclasses come first
        for ci in self.converters:
            if issubclass(parameter.type, ci.convert_type):
                return ci.convert_func.__get__(self, self.__class__)(value, parameter, task)

        # if we get here, we don't have a converter for this type
        if parameter.type is inspect.Parameter.empty:
            raise ValueError('Cannot convert parameter named {} because it has no type hint and CONVERT_WITHOUT_TYPE_HINTS is True'.format(parameter.name))
        raise ValueError('No parameter converter exists for type: {}'.format(parameter.type))



class DefaultConverter(BaseConverter):
    '''
    The default converter that comes with DMP.  URL parameter converters
    can be any callable.  This implementation is an extensible class with pluggable
    methods for conversion of different types.

    Converters are any type of callable; see __call__ below for the primary method.
    '''
    # characters that mean None values in URLs
    EMPTY_CHARACTERS = { '', '-', '0', None }

    @BaseConverter.convert_method(str)
    def convert_str(self, value, parameter, task):
        '''Pass through for strings'''
        return value


    @BaseConverter.convert_method(int, float)
    def convert_number(self, value, parameter, task):
        '''Converts the string to an int or float'''
        if value is None:
            value = 0
        try:
            return parameter.type(value)
        except Exception as e:
            log.info('Raising Http404 due to parameter conversion error: %s', e)
            raise Http404('Invalid parameter specified in the url')


    @BaseConverter.convert_method(decimal.Decimal)
    def convert_decimal(self, value, parameter, task):
        '''Converts the string to a decimal.Decimal'''
        if value in self.EMPTY_CHARACTERS:
            return None
        try:
            return parameter.type(value)
        except Exception as e:
            log.info('Raising Http404 due to parameter conversion error: %s', e)
            raise Http404('Invalid parameter specified in the url')


    @BaseConverter.convert_method(bool)
    def convert_boolean(self, value, parameter, task):
        '''Converts the string to float'''
        try:
            return value not in self.EMPTY_CHARACTERS
        except Exception as e:
            log.info('Raising Http404 due to parameter conversion error: %s', e)
            raise Http404('Invalid parameter specified in the url')


    @BaseConverter.convert_method(datetime.datetime)
    def convert_datetime(self, value, parameter, task):
        '''Converts the string to datetime.datetime'''
        if value in self.EMPTY_CHARACTERS:
            return None
        try:
            for fmt in settings.DATETIME_INPUT_FORMATS:
                try:
                    return datetime.datetime.strptime(value, fmt)
                except (ValueError, TypeError):
                    continue
            raise ValueError("value '{}' does not match any formats in settings.DATETIME_INPUT_FORMATS".format(value))
        except Exception as e:
            log.info('Raising Http404 due to parameter conversion error: %s', e)
            raise Http404('Invalid parameter specified in the url')


    @BaseConverter.convert_method(datetime.date)
    def convert_date(self, value, parameter, task):
        '''Converts the string to datetime.date'''
        if value in self.EMPTY_CHARACTERS:
            return None
        try:
            for fmt in settings.DATE_INPUT_FORMATS:
                try:
                    return datetime.datetime.strptime(value, fmt).date()
                except (ValueError, TypeError):
                    continue
            raise ValueError("value '{}' does not match any formats in settings.DATE_INPUT_FORMATS".format(value))
        except Exception as e:
            log.info('Raising Http404 due to parameter conversion error: %s', e)
            raise Http404('Invalid parameter specified in the url')


    @BaseConverter.convert_method(Model)  # django models.Model
    def convert_id_to_model(self, value, parameter, task):
        '''
        Converts a urlparam id to a model object.
            - The primary key (id) is expected to be an int (standard django way of doing it).
            - An empty string, dash "-", or 0 returns None.
            - Anything else raises Http404, including a DoesNotExist on the .get() call.
        '''
        if value in self.EMPTY_CHARACTERS:
            return None
        try:
            pk = int(value)
        except Exception as e:
            log.info('Raising Http404 due to parameter conversion error: %s', e)
            raise Http404('Invalid parameter specified in the url')
        try:
            return parameter.type.objects.get(id=pk)
        except ObjectDoesNotExist as e:
            log.info('Raising Http404 due to parameter conversion error: %s', e)
            raise Http404('Invalid parameter specified in the url')



####################################################
###   Setting and getting of default converter

DEFAULT_CONVERTER_KEY = '_dmp_default_converter'

def _check_converter(converter):
    if converter is None:
        return get_default_converter()
    elif inspect.isclass(converter):
        if hasattr(converter, '__call__'):
            return converter()
        else:
            raise ValueError('Converters must be callable or classes that implements the __call__ method.')
    elif callable(converter):
        return converter
    raise ValueError('Converters must be callable or classes that implements the __call__ method.')


def set_default_converter(converter=DefaultConverter):
    '''
    Sets the default converter used for view function parameters.

    `converter` should be a callable or a class with a __call__ method.
    If None, the system returns to the default DMP converter.
    '''
    DMP_OPTIONS[DEFAULT_CONVERTER_KEY] = _check_converter(converter)


def get_default_converter():
    '''
    Returns the default converter in the DMP system.
    '''
    return DMP_OPTIONS[DEFAULT_CONVERTER_KEY]


# set the intiial converter for the system
set_default_converter()


from django.conf import settings
from django.apps import apps
from django.template import TemplateDoesNotExist

from mako.exceptions import MakoException
import mako.runtime

from .sass import check_template_scss
from .exceptions import SassCompileException
from .util import get_dmp_instance, log, DMP_OPTIONS

import os, os.path, io, posixpath, warnings
from collections import deque




# keys to attach TemplateInfo objects and static renderer to the Mako Template itself
# I attach it to the template objects because they are already cached by mako, so
# caching them again here would result in double-caching.  It's a bit of a
# monkey-patch to set them as attributes in the Mako templates, but it's efficient
# and encapsulated entirely within this file
DMP_TEMPLATEINFO_KEY = '_django_mako_plus_templateinfo'

# a cache to store TemplateInfo objects that didn't have an associated template
# these are normally nonexistent files that were referenced (which we allow)
NO_TSELF_CACHE = {}


#######################################################################
###   Shortcut methods - these are the primary way the static render
###   should be used


def link_css(tself, cgi_id=None):
    '''
    Renders the <style> links for a given template.

        ${ django_mako_plus.link_css(self) }

    This call is normally placed in the <head> section of your base.htm
    template.

    Suppose you have two html files using template inheritance: index.html
    and base.htm.  This method will render the links for all style files
    with the names of either of these two templates.  Up to four files
    are linked/sourced:

        base.htm       Generates a <style> link for app/styles/base.css
            |          and a <style> source for app/styles/base.cssm
            |
        index.html     Generates a <style> link for app/styles/index.css
                       and a <style> source for app/styles/index.cssm

    This call must be made from within a rendering Mako template because
    this is where you have  access to the "self" namespace.

    If you need to render these links outside of a template, see link_template_css()
    below.

    The optional cgi_id parameter is to overcome browser caches.  On some browsers,
    changes to your CSS/JS files don't get downloaded because the browser waits a time
    to check for a new version.  This wait time is set by your web server,
    and it's normally a good thing to speed everything up.  However,
    when you upload new CSS/JS files, you want all browsers to download the new
    files even if their cached versions have't expired yet.
    By adding an arbitrary id to the end of the .css and .js files, browsers will
    see the files as *new* anytime that id changes.  The default method
    for calculating the id is the file modification time (minutes since 1970).
    '''
    html = []
    for ti in reversed(build_templateinfo_chain(tself, cgi_id)):
        ti.append_css(tself.context.get('request'), tself.context, html)
    return '\n'.join(html)


def link_js(tself, cgi_id=None):
    '''
    Renders the <script> links for a given template:

        ${ django_mako_plus.link_js(self) }

    This call is normally placed near the end of your base.htm
    template.

    Suppose you have two html files using template inheritance: index.html
    and base.htm.  This method will render the links for all script files
    with the names of either of these two templates.  Up to four files
    are linked/sourced:

        base.htm       Generates a <script> link for app/scripts/base.js
            |          and a <script> source for app/scripts/base.jsm
            |
        index.html     Generates a <script> link for app/scripts/index.js
                       and a <script> source for app/scripts/index.jsm

    This call must be made from within a rendering Mako template because
    this is where you have  access to the "self" namespace.

    If you need to render these links outside of a template, see link_template_js()
    below.

    The optional cgi_id parameter is to overcome browser caches.  On some browsers,
    changes to your CSS/JS files don't get downloaded because the browser waits a time
    to check for a new version.  This wait time is set by your web server,
    and it's normally a good thing to speed everything up.  However,
    when you upload new CSS/JS files, you want all browsers to download the new
    files even if their cached versions have't expired yet.
    By adding an arbitrary id to the end of the .css and .js files, browsers will
    see the files as *new* anytime that id changes.  The default method
    for calculating the id is the file modification time (minutes since 1970).
    '''
    html = []
    for ti in reversed(build_templateinfo_chain(tself, cgi_id)):
        ti.append_js(tself.context.get('request'), tself.context, html)
    return '\n'.join(html)


def link_template_css(request, app, template_name, context, cgi_id=None, force=True):
    '''
    Renders the styles for the given template in the given app.  Normally,
    link_css() is used to accomplish this in your base template.  This method
    is available when/if you need to generate the links in normal python
    code (rather than within a running template).

    The app can be an AppConfig object or the name of the app.
    The template_name is the name of a file in the app/templates/ directory.

    The related .css/.cssm files are created for the template_name
    even if a nonexistent filename is sent.  This can be useful when
    creating html directly within Python code.
    '''
    html = []
    for ti in reversed(build_templateinfo_chain_by_name(app, template_name, cgi_id, force)):
        ti.append_css(request, context, html)
    return '\n'.join(html)


def link_template_js(request, app, template_name, context, cgi_id=None, force=True):
    '''
    Renders the scripts for the given template in the given app.  Normally,
    link_js() is used to accomplish this in your base template.  This method
    is available when/if you need to generate the links in normal python
    code (rather than within a running template).

    The app can be an AppConfig object or the name of the app.
    The template_name is the name of a file in the app/templates/ directory.

    The related .js/.jsm files are created for the template_name
    even if a nonexistent filename is sent.  This can be useful when
    creating html directly within Python code.
    '''
    html = []
    for ti in reversed(build_templateinfo_chain_by_name(app, template_name, cgi_id, force)):
        ti.append_js(request, context, html)
    return '\n'.join(html)



###########################################################
###   Deprecated as of Jan 2017

def get_template_css(tself, request, context, cgi_id=None):
    warnings.warn('As of DMP 3.8, get_template_css(self, request, context) has been changed to template_css(self).  Please make the necessary adjustments in your base templates.', DeprecationWarning)
    return link_css(tself, cgi_id)


def get_template_js(tself, request, context, cgi_id=None):
    warnings.warn('As of DMP 3.8, get_template_js(self, request, context) has been changed to template_js(self).  Please make the necessary adjustments in your base templates.', DeprecationWarning)
    return link_js(tself, cgi_id)




##############################################################################
###   Builds a chain of TemplateInfo objects, one of each level
###   of the inheritance chain of a template.
###


def build_templateinfo_chain(tself, cgi_id=None):
    '''
    Retrieves a chain of TemplateInfo objects.  The chain is formed by following
    template inheritance through inherit tags: <%inherit file="base_homepage.htm" />

    For efficiency, the templates are ordered with the specialization template first.
    The returned list should usually be reversed() by the calling code to put the
    superclass CSS/JS first (allowing the specialization to override supertemplate styles
    and scripts).

    This function is not normally used directly.  Use the link_css() and link_js()
    functions instead.

    # Example placed in a template, using the self namespace
        %for ti in build_templateinfo_chain(self):
            ...
        %endfor

    '''
    # making a list (instead of generator) because it gets reversed() in functions like link_css()
    chain = []

    # step through the template inheritance
    while tself is not None:
        # first check the cache, creating if necessary
        ti = getattr(tself.template, DMP_TEMPLATEINFO_KEY, None)
        if ti is None:
            template_dir, template_name = os.path.split(tself.template.filename)
            app_dir = os.path.dirname(template_dir)
            ti = TemplateInfo(app_dir, template_name, cgi_id)
            # cache in template if in production mode
            if not settings.DEBUG:
                setattr(tself.template, DMP_TEMPLATEINFO_KEY, ti)
        # append the TemplateInfo
        chain.append(ti)
        # loop with the next inherited template
        tself = tself.inherits

    # return
    return chain


def build_templateinfo_chain_by_name(app, template_name, cgi_id, force=True):
    '''
    Retrieves a chain of TemplateInfo objects.  The chain is formed by following
    template inheritance through inherit tags: <%inherit file="base_homepage.htm" />

    For efficiency, the templates are ordered with the specialization template first.
    The returned list should usually be reversed() by the calling code to put the
    superclass CSS/JS first (allowing the specialization to override supertemplate styles
    and scripts).

    This version of the function is useful for python-generated html
    (i.e. direct view.py output).

    If force is True, the templateinfo chain will be created even if the template_name
    does not exist.  This allows python-generated code to take advantage of DMP's
    automatic CSS/JS rendering even if the code has no template.

    If force is False, Mako errors are raised as they occur.  See:
        mako.exceptions.CompileException
        mako.exceptions.SyntaxException
        mako.exceptions.TemplateLookupException
        mako.exceptions.MakoException (catch-all for all Mako errors)

    # Example placed in any python code:
        for ti in build_templateinfo_chain_by_name('homepage', 'index.html'):
            ...

    '''
    # get the AppConfig
    if isinstance(app, str):
        app = apps.get_app_config(app)

    # try to generate a tself so we can use the normal method
    # this uses Mako caching mechanism, so it's pretty fast
    try:
        template = get_dmp_instance().get_template_loader(app, create=True).get_mako_template(template_name)
        context = _create_empty_mako_context(template)
        return build_templateinfo_chain(context['self'], cgi_id)
    except MakoException: # includes template not found, template syntax error, compile exception, etc.
        if not force:
            raise

    # if we get here, we probably have a nonexistent file (which is allowed)
    key = (app.path, template_name)
    try:
        ti = NO_TSELF_CACHE[key]
    except KeyError:
        ti = TemplateInfo(app.path, template_name, cgi_id)
        # cache if in production mode
        if not settings.DEBUG:
            NO_TSELF_CACHE[key] = ti
    return [ ti ]


class TemplateInfo(object):
    '''
    Data class that holds information about a template's directories.  A TemplateInfo object
    is created by build_templateinfo_chain for each level in the Mako template inheritance chain.
    This object is then attached to the template object, which Mako already caches.
    That way we only do this work once per server run (in production mode).

    The app_dir is the search location for the styles/ and scripts/ folders, relative to the
    project base directory.  For typical apps, this is simply the app name.

    The template_name is the filename of the template.

    The cgi_id parameter is described in the MakoTemplateRenderer class below.
    '''
    def __init__(self, app_dir, template_name, cgi_id=None):
        # calculate directory and static url locations
        self.template_name = os.path.splitext(template_name)[0]  # remove its extension
        self.app_dir = app_dir
        app_reldir = os.path.relpath(self.app_dir, settings.BASE_DIR)
        self.app_url = posixpath.join(*app_reldir.split(os.path.sep))  # ensure we have forward slashes (even on windwos) because this is for urls

        # set up the filenames
        css_file = os.path.join(self.app_dir, 'styles', '%s.css' % self.template_name)
        cssm_file = os.path.join(self.app_dir, 'styles', '%s.cssm' % self.template_name)
        js_file = os.path.join(self.app_dir, 'scripts', '%s.js' % self.template_name)
        jsm_file = os.path.join(self.app_dir, 'scripts', '%s.jsm' % self.template_name)

        # the SASS templatename.scss (compile any updated templatename.scss files to templatename.css files)
        if DMP_OPTIONS.get('RUNTIME_SCSS_ENABLED'):
            check_template_scss(os.path.join(self.app_dir, 'styles'), self.template_name)

        # I want short try blocks, so there are several - for example, the first OSError can only occur for one reason: if the os.stat() fails.
        # I'm using os.stat here instead of os.path.exists because I need the st_mtime.  The os.stat checks that the file exists and gets the modified time in one command.

        # the static templatename.css file
        try:
            fstat = os.stat(css_file)
            self.css = '<link rel="stylesheet" type="text/css" href="%s?%s" />' % (posixpath.join(settings.STATIC_URL, self.app_url, 'styles', self.template_name + '.css'), cgi_id if cgi_id != None else int(fstat.st_mtime))
        except OSError:
            self.css = None

        # the mako-rendered templatename.cssm file
        try:
            fstat = os.stat(cssm_file)
            self.cssm = self.template_name + '.cssm'
        except OSError:
            self.cssm = None

        # the static templatename.js file
        try:
            fstat = os.stat(js_file)
            self.js = '<script src="%s?%s"></script>' % (posixpath.join(settings.STATIC_URL, self.app_url, 'scripts', self.template_name + '.js'), cgi_id if cgi_id != None else int(fstat.st_mtime))
        except OSError:
            self.js = None

        # the mako-rendered templatename.jsm file
        try:
            fstat = os.stat(jsm_file)
            self.jsm = self.template_name + '.jsm'
        except OSError:
            self.jsm = None


    def append_css(self, request, context, html):
        '''Appends the CSS for this template's .css and .cssm files, if they exist, to the html list.'''
        # do we have a css?
        if self.css:
            html.append(self.css)  # the <link> was already created once in the constructor
        # do we have a cssm?
        if self.cssm:
            # engine.py already caches these loaders, so no need to cache them again here
            lookup = get_dmp_instance().get_template_loader_for_path(os.path.join(self.app_dir, 'styles'))
            css_text = lookup.get_template(self.cssm).render(request=request, context=context)
            if DMP_OPTIONS.get('RUNTIME_CSSMIN'):
                css_text = DMP_OPTIONS['RUNTIME_CSSMIN'](css_text)
            html.append('<style type="text/css">%s</style>' % css_text)


    def append_js(self, request, context, html):
        '''Appends the Javascript for this template's .js and .jsm files, if they exist to the html list.'''
        # do we have a js?
        if self.js:
            html.append(self.js)  # the <script> was already created once in the constructor
        # do we have a jsm?
        if self.jsm:
            # engine.py already caches these loaders, so no need to cache them again here
            lookup = get_dmp_instance().get_template_loader_for_path(os.path.join(self.app_dir, 'scripts'))
            js_text = lookup.get_template(self.jsm).render(request=request, context=context)
            if DMP_OPTIONS.get('RUNTIME_JSMIN'):
                js_text = DMP_OPTIONS['RUNTIME_JSMIN'](js_text)
            html.append('<script>%s</script>' % js_text)




#######################################################################
###   Utility functions

def _create_empty_mako_context(template):
    '''
    A small utility function that generates a Mako Context, used to
    get the inheritance for a template in build_templateinfo_chain()  above.
    This is only used when app and template_name are used.

    I'm hacking into the Mako private functions because I don't see a better way
    to get template inheritance.  This utility function at least contains it in
    one place.
    '''
    context = mako.runtime.Context(io.BytesIO())
    context._set_with_template(template)
    func, lclcontext = mako.runtime._populate_self_namespace(context, template)
    return lclcontext





from django.core.exceptions import ImproperlyConfigured
from django.views.generic import View
from urllib.parse import unquote

# try to import MiddlewareMixIn (Django 1.10+)
try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    # create a dummy MiddlewareMixin if older Django
    MiddlewareMixin = object

from .router import get_router, route_request, ClassBasedRouter
from .util import URLParamList, get_dmp_instance, DMP_OPTIONS, log

import logging

##########################################################
###   Middleware the prepares the request for
###   use with the controller.

EMPTY_URLPARAMS = URLParamList()
DMP_PARAM_CHECK = ( 'dmp_router_app', 'dmp_router_page', 'dmp_router_function', 'urlparams' )

class RequestInitMiddleware(MiddlewareMixin):
    '''
    Adds several fields to the request that our controller needs.
    Projects can customize the variables with view middleware that runs after this class.

    This class must be included in settings.py -> MIDDLEWARE.
    '''
    def process_request(self, request):
        '''
        Adds stubs for the DMP custom variables to the request object.

        We do this early in the middleware process because process_view() below gets called
        *after* URL resolution.  If an exception occurs before that (a 404, a middleware exception),
        the variables won't be put on the request object.  Therefore, we put stubs on it at the
        earliest possible place so they exist even if we don't get to process_view().
        '''
        request.dmp_router_app = None
        request.dmp_router_page = None
        request.dmp_router_function = None
        request.dmp_router_module = None
        request.dmp_router_class = None
        request.urlparams = EMPTY_URLPARAMS
        request._dmp_router_callable = None


    def process_view(self, request, view_func, view_args, view_kwargs):
        '''
        Adds the following to the request object:

            request.dmp_router_app        The Django application (such as "homepage"), as a string.
            request.dmp_router_page       The view module (such as "index" for index.py), as a string.
            request.dmp_router_function   The function within the view module to be called (usually "process_request"), as a string.
            request.dmp_router_module     The module path in Python terms (such as homepage.views.index), as a string.
            request.dmp_router_class      This is set to None in this method, but route_request() fills it in, as a string, if a class-based view.
            request.urlparams             A list of the remaining url parts, as a list of strings.  See the tutorial for more information on this parameter.
            request._dmp_router_callable   The view callable (function, method, etc.) to be called by the router.

        Named parameters in the url pattern determines the values of these variables.
        See django_mako_plus/urls.py for the DMP standard patterns.  For example, one pattern
        specifies the router app, page, function, and urlparams: /app/page.function/urlparams

            ^(?P<dmp_router_app>[_a-zA-Z0-9\-]+)/(?P<dmp_router_page>[_a-zA-Z0-9\-]+)\.(?P<dmp_router_function>[_a-zA-Z0-9\.\-]+)/?(?P<urlparams>.*)$

        As view middleware, this function runs just before the router.route_request is called.
        '''
        # set a flag on the request to check for double runs (middleware listed twice in settings)
        # double runs don't work because we pop off the kwargs below
        if hasattr(request, '_dmp_router_middleware_flag'):
            raise ImproperlyConfigured('The Django Mako Plus middleware is running twice on this request.  Please check settings.py to see it the middleware might be listed twice.')
        request._dmp_router_middleware_flag = True

        # print debug messages to help with urls.py regex issues
        if log.isEnabledFor(logging.DEBUG):
            kwarg_params = [ param for param in DMP_PARAM_CHECK if param in view_kwargs ]
            missing_params = [ param for param in DMP_PARAM_CHECK if param not in view_kwargs ]
            log.debug('variables set by urls.py: %s; variables set by defaults: %s', kwarg_params, missing_params)

        # add the variables to the request
        request.dmp_router_app = view_kwargs.pop('dmp_router_app', None) or DMP_OPTIONS.get('DEFAULT_APP', 'homepage')
        request.dmp_router_page = view_kwargs.pop('dmp_router_page', None) or DMP_OPTIONS.get('DEFAULT_PAGE', 'index')
        request.dmp_router_function = view_kwargs.pop('dmp_router_function', None)
        if request.dmp_router_function:
            fallback_template = '{}.{}.html'.format(request.dmp_router_page, request.dmp_router_function)
        else:
            fallback_template = '{}.html'.format(request.dmp_router_page)
            request.dmp_router_function = 'process_request'

        # period and dash cannot be in python names, but we allow dash in app and page and (dash or period) in the function
        # these change into underscores
        request.dmp_router_app = request.dmp_router_app.replace('-', '_')
        request.dmp_router_page = request.dmp_router_page.replace('-', '_')
        request.dmp_router_function = request.dmp_router_function.replace('.', '_').replace('-', '_')

        # add the full module path to the request
        request.dmp_router_module = '.'.join([ request.dmp_router_app, 'views', request.dmp_router_page ])

        # add the url parameters to the request
        # note that I'm not using unquote_plus because the + switches to a space *after* the question mark (in the regular parameters)
        # in the normal url, spaces should be quoted with %20.  Thanks Rosie for the tip.
        kwarg_urlparams = view_kwargs.pop('urlparams', '').strip()
        if kwarg_urlparams:
            request.urlparams = URLParamList(( unquote(s) for s in kwarg_urlparams.split('/') ))
        else:
            request.urlparams = URLParamList()

        # get the function object - the return of get_router_function might be a function, a class-based view, or a template
        # get_router_function does some magic to make all of these act like a regular view function
        request._dmp_router_callable = get_router(request.dmp_router_module, request.dmp_router_function, request.dmp_router_app, fallback_template)

        # adjust the variable values if a class
        if isinstance(request._dmp_router_callable, ClassBasedRouter):
            request.dmp_router_class = request.dmp_router_function
            request.dmp_router_function = request.method.lower()

        # debugging
        # print('request.dmp_router_app        ', request.dmp_router_app        )
        # print('request.dmp_router_page       ', request.dmp_router_page       )
        # print('request.dmp_router_function   ', request.dmp_router_function   )
        # print('request.dmp_router_module     ', request.dmp_router_module     )
        # print('request.dmp_router_class      ', request.dmp_router_class      )
        # print('request.urlparams             ', request.urlparams             )
        # print('request._dmp_router_callable  ', request._dmp_router_callable  )



from mako.lexer import Lexer
from mako import parsetree

from .exceptions import SassCompileException
from .util import DMP_OPTIONS, run_command, lock_file, encode32, decode32

import glob, sys, os, os.path, subprocess, itertools, re


__doc__ = '''
Compiles scss files using the Sass runtime arguments in settings.py.

Also compiles scssm files, which are scss files with embedded Mako.
Only ${ var } expressions are supported right now.  Other Mako
constructs like <% %> may be supported later.
'''


MAKO_EXPRESSION_MARKERS = ( 'MakoExpression', 'ExpressionMako' )
RE_MAKO_EXPRESSION = re.compile('{}_(.+?)_{}'.format(MAKO_EXPRESSION_MARKERS[0], MAKO_EXPRESSION_MARKERS[1]), flags=re.MULTILINE)


def check_template_scss(styles_dir, template_name):
    '''
    Checks to see if template.scss or template.scssm need compiling.
    The styles_dir should be the full path to the styles directory.
    The template_name should be the name of the template *without the extension*.

    Raises a SassCompileException if compilation fails.
    '''
    for ext in ( 'css', 'cssm' ):
        scss_file = os.path.join(styles_dir, '{}.s{}'.format(template_name, ext))
        gen_css_file = os.path.join(styles_dir, '{}.{}'.format(template_name, ext))
        try:
            scss_stat = os.stat(scss_file)
        except OSError:
            scss_stat = None
        if scss_stat != None:  # only continue this block if we found a .scss file
            try:
                fstat = os.stat(gen_css_file)
            except OSError:
                fstat = None
            # if we 1) have no css_file or 2) have a newer scss_file, run the compiler
            if fstat == None or scss_stat.st_mtime > fstat.st_mtime:
                try:
                    if ext == 'css':
                        compile_scss_file(scss_file, gen_css_file)
                    else:
                        compile_scssm_file(scss_file, gen_css_file)
                except subprocess.CalledProcessError as cpe:
                    raise SassCompileException(cpe.cmd, cpe.stderr)


def compile_scss_file(scss_file, css_file):
    '''
    Compiles a regular scss file

    If the Sass execution fails, this function raises subprocess.CalledProcessError.
    '''
    # run sass on it
    args = list(DMP_OPTIONS.get('RUNTIME_SCSS_ARGUMENTS'))
    args.append(scss_file)
    args.append(css_file)
    run_command(*args)


def compile_scssm_file(scssm_file, css_file):
    '''
    Compiles an scssm (sass with embedded Mako) file.
    This is a total hack to sneak mako codes through the Sass compiler.  The algorithm
    uses the Mako lexer to split the file, then it replaces Expressions in the file
    with Base32 -- that alphabet is valid throughout CSS rules.  After Sass does its thing,
    we switch the Base32 back to the Mako codes.

    If the Sass execution fails, this function raises subprocess.CalledProcessError.
    '''
    # this algorithm isn't terribly fast, but it only runs the first time the file is accessed on a production system
    # lock the file so other processes can't access it
    encoding = settings.DEFAULT_CHARSET or 'utf8'
    with lock_file(scssm_file):

        # read the contents of the file
        with open(scssm_file, encoding=encoding) as fin:
            contents = fin.read()

        try:
            # make a backup just in case something goes wrong (this is for testing)
            # backup_scss = '{}.bak'.format(scssm_file)
            # with open(backup_scss, 'w', encoding=encoding) as fout:
            #     fout.write(contents)
            # embed mako codes in markers that are acceptable to sass
            with open(scssm_file, 'w', encoding=encoding) as fout:
                for node in Lexer(text=contents).parse().nodes:
                    if isinstance(node, parsetree.Expression):
                        fout.write('{}_{}_{}'.format(MAKO_EXPRESSION_MARKERS[0], encode32(node.text), MAKO_EXPRESSION_MARKERS[1]))
                    elif isinstance(node, parsetree.Text):
                        fout.write(node.content)
            # run sass on it
            compile_scss_file(scssm_file, css_file)
            # read the contents of the generated file
            with open(css_file, encoding=encoding) as fin:
                csscontents = fin.read()
            # unencode all the mako expressions
            csscontents = RE_MAKO_EXPRESSION.sub(lambda match: '${{{}}}'.format(decode32(match.group(1))), csscontents)
            # write the contents
            with open(css_file, 'w', encoding=encoding) as fout:
                fout.write(csscontents)
            # erase the backup
            # os.remove(backup_scss)

        finally:
            # replace the original contents back in the scss file
            with open(scssm_file, 'w', encoding=encoding) as fout:
                fout.write(contents)
            # update the modified timestamp on the css file so the scssm_file is older than it (we just rewrote the scssm)
            os.utime(css_file)





from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from django_mako_plus.util import get_dmp_instance, get_dmp_app_configs, DMP_OPTIONS

from optparse import make_option
import os, os.path, shutil, fnmatch
from importlib import import_module



TYPE_DIRECTORY = 0       # file is a directory
TYPE_FILE = 1            # file is a regular file


class Rule(object):
    def __init__(self, pattern, level, filetype, score):
        self.pattern = pattern
        self.level = level
        self.filetype = filetype
        self.score = score

    def match(self, fname, flevel, ftype):
        '''Returns the result score if the file matches this rule'''
        # if filetype is the same
        # and level isn't set or level is the same
        # and pattern matche the filename
        if self.filetype == ftype and (self.level is None or self.level == flevel) and fnmatch.fnmatch(fname, self.pattern):
            return self.score
        return 0


INITIAL_RULES = (
    # files are included by default
    Rule('*',                                    level=None, filetype=TYPE_FILE,      score=1),
    # files at the app level are skipped
    Rule('*',                                    level=0,    filetype=TYPE_FILE,      score=-2),
    # directories are recursed by default
    Rule('*',                                    level=None, filetype=TYPE_DIRECTORY, score=1),
    # directories at the app level are skipped
    Rule('*',                                    level=0,    filetype=TYPE_DIRECTORY, score=-2),

    # media, scripts, styles directories are what we want to copy
    Rule('media',                                level=0,    filetype=TYPE_DIRECTORY, score=6),
    Rule('scripts',                              level=0,    filetype=TYPE_DIRECTORY, score=6),
    Rule('styles',                               level=0,    filetype=TYPE_DIRECTORY, score=6),

    # ignore the template cache directories
    Rule(DMP_OPTIONS.get('TEMPLATES_CACHE_DIR'), level=None, filetype=TYPE_DIRECTORY, score=-3),
    # ignore python cache directories
    Rule('__pycache__',                          level=None, filetype=TYPE_DIRECTORY, score=-3),
    # ignore compiled python files
    Rule('*.pyc',                                level=None, filetype=TYPE_FILE,      score=-3),
    # ignore all cssm and jsm files (these are rendered at runtime, so not static)
    Rule('*.cssm',                               level=None, filetype=TYPE_FILE,      score=-3),
    Rule('*.jsm',                                level=None, filetype=TYPE_FILE,      score=-3),
)




# import minification if requested
JSMIN = False
CSSMIN = False
if DMP_OPTIONS.get('MINIFY_JS_CSS', False):
    try:
        from rjsmin import jsmin
        JSMIN = True
    except ImportError:
        raise CommandError('The Django Mako Plus option "MINIFY_JS_CSS" is True in settings.py, but the "rjsmin" module does not seem to be installed. Do you need to "pip install" it?')
    try:
        from rcssmin import cssmin
        CSSMIN = True
    except ImportError:
        raise CommandError('The Django Mako Plus option "MINIFY_JS_CSS" is True in settings.py, but the "rcssmin" module does not seem to be installed. Do you need to "pip install" it?')



class Command(BaseCommand):
    args = ''
    help = 'Collects static files, such as media, scripts, and styles, to a common directory root. This is done to prepare for deployment.'
    can_import_settings = True


    def add_arguments(self, parser):
        parser.add_argument(
            '--overwrite',
            action='store_true',
            dest='overwrite',
            default=False,
            help='Overwrite existing files in the directory when necessary.'
        )
        parser.add_argument(
            '--verbose',
            action='store_true',
            dest='verbose',
            default=False,
            help='Set verbosity to level 3 (see --verbosity).'
        )
        parser.add_argument(
            '--quiet',
            action='store_true',
            dest='quiet',
            default=False,
            help='Set verbosity to level 0, which silences all messages (see --verbosity).'
        )
        parser.add_argument(
            '--include-dir',
            action='append',
            dest='include_dir',
            help='Include directories matching this pattern.  Unix-style wildcards are acceptable, such as "*partial*".  This option can be specified more than once.'
        )
        parser.add_argument(
            '--include-file',
            action='append',
            dest='include_file',
            help='Include files matching this pattern.  Unix-style wildcards are acceptable, such as "*.txt".  This option can be specified more than once.'
        )
        parser.add_argument(
            '--skip-dir',
            action='append',
            dest='skip_dir',
            help='Skip directories matching this pattern.  Unix-style wildcards are acceptable, such as "*partial*".  This option can be specified more than once.'
        )
        parser.add_argument(
            '--skip-file',
            action='append',
            dest='skip_file',
            help='Skip files matching this pattern.  Unix-style wildcards are acceptable, such as "*.txt".  This option can be specified more than once.'
        )


    def handle(self, *args, **options):
        # save the options for later
        self.options = options
        if self.options['verbose']:
            self.options['verbosity'] = 3
        if self.options['quiet']:
            self.options['verbosity'] = 0

        # ensure we have a base directory
        try:
            if not os.path.isdir(os.path.abspath(settings.BASE_DIR)):
                raise CommandError('Your settings.py BASE_DIR setting is not a valid directory.  Please check your settings.py file for the BASE_DIR variable.')
        except AttributeError as e:
            print(e)
            raise CommandError('Your settings.py file is missing the BASE_DIR setting. Aborting app creation.')

        # get the destination directory, and ensure it doesn't already exist
        # if dest_root starts with a /, it is an absolute directory
        # if dest_root doesn't start with a /, it goes relative to the BASE_DIR
        try:
            dest_root = os.path.join(os.path.abspath(settings.BASE_DIR), settings.STATIC_ROOT)
            if os.path.exists(dest_root) and not self.options['overwrite']:
                raise CommandError('The destination directory for static files (%s) already exists. Please delete it or run this command with the --overwrite option.' % dest_root)
        except AttributeError:
            raise CommandError('Your settings.py file is missing the STATIC_ROOT setting. Exiting without collecting the static files.')

        # create the directory - we assume it either doesn't exist, or the --overwrite is specified
        if not os.path.isdir(dest_root):
            os.makedirs(dest_root)

        # set up the rules
        self.rules = list(INITIAL_RULES)
        self.add_user_rules()

        # go through the DMP apps and collect the static files
        for config in get_dmp_app_configs():
            self.message('Processing app {}'.format(config.name), 1)
            self.copy_dir(config.path, os.path.abspath(os.path.join(dest_root, config.name)))


    def message(self, msg, level, tab=0):
        '''Print a message to the console'''
        # verbosity=1 is the default if not specified in the options
        if self.options['verbosity'] >= min(level, 3):
            print('{}{}'.format('    ' * tab, msg))


    def add_user_rules(self):
        '''Adds rules for the command line options'''
        # include rules have score of 50 because they trump all initial rules
        for pattern in (self.options['include_dir'] or []):
            self.message('Setting rule - recurse directories: {}'.format(pattern), 1)
            self.rules.append(Rule(pattern, level=None, filetype=TYPE_DIRECTORY, score=50))
        for pattern in (self.options['include_file'] or []):
            self.message('Setting rule - include files: {}'.format(pattern), 1)
            self.rules.append(Rule(pattern, level=None, filetype=TYPE_FILE, score=50))
        # skip rules have score of 100 because they trump everything, including the includes from the command line
        for pattern in (self.options['skip_dir'] or []):
            self.message('Setting rule - skip directories: {}'.format(pattern), 1)
            self.rules.append(Rule(pattern, level=None, filetype=TYPE_DIRECTORY, score=-100))
        for pattern in (self.options['skip_file'] or []):
            self.message('Setting rule - skip files: {}'.format(pattern), 1)
            self.rules.append(Rule(pattern, level=None, filetype=TYPE_FILE, score=-100))


    def copy_dir(self, source, dest, level=0):
        '''Copies the static files from one directory to another.  If this command is run, we assume the user wants to overwrite any existing files.'''
        encoding = settings.DEFAULT_CHARSET or 'utf8'
        msglevel = 2 if level == 0 else 3
        self.message('Directory: {}'.format(source), msglevel, level)

        # create a directory for this app
        if not os.path.exists(dest):
            self.message('Creating directory: {}'.format(dest), msglevel, level+1)
            os.mkdir(dest)

        # go through the files in this app
        for fname in os.listdir(source):
            source_path = os.path.join(source, fname)
            dest_path = os.path.join(dest, fname)
            ext = os.path.splitext(fname)[1].lower()

            # get the score for this file
            score = 0
            for rule in self.rules:
                score += rule.match(fname, level, TYPE_DIRECTORY if os.path.isdir(source_path) else TYPE_FILE)

            # if score is not above zero, we skip this file
            if score <= 0:
                self.message('Skipping file with score {}: {}'.format(score, source_path), msglevel, level+1)
                continue

            ### if we get here, we need to copy the file ###

            # if a directory, recurse to it
            if os.path.isdir(source_path):
                self.message('Creating directory with score {}: {}'.format(score, source_path), msglevel, level+1)
                # create it in the destination and recurse
                if not os.path.exists(dest_path):
                    os.mkdir(dest_path)
                elif not os.path.isdir(dest_path):  # could be a file or link
                    os.unlink(dest_path)
                    os.mkdir(dest_path)
                self.copy_dir(source_path, dest_path, level+1)

            # if a regular Javscript file, minify it
            elif ext == '.js' and DMP_OPTIONS.get('MINIFY_JS_CSS', False) and JSMIN:
                self.message('Including and minifying file with score {}: {}'.format(score, source_path), msglevel, level+1)
                with open(source_path, encoding=encoding) as fin:
                    with open(dest_path, 'w', encoding=encoding) as fout:
                        minified = minify(fin.read(), jsmin)
                        fout.write(minified)


            # same with css files
            elif ext == '.css' and DMP_OPTIONS.get('MINIFY_JS_CSS', False) and CSSMIN:
                self.message('Including and minifying file with score {}: {}'.format(score, source_path), msglevel, level+1)
                with open(source_path, encoding=encoding) as fin:
                    with open(dest_path, 'w', encoding=encoding) as fout:
                        minified = minify(fin.read(), cssmin)
                        fout.write(minified)

            # otherwise, just copy the file
            else:
                self.message('Including file with score {}: {}'.format(score, source_path), msglevel, level+1)
                shutil.copy2(source_path, dest_path)



##############################################
###   Utility functions

def minify(text, minifier):
    '''Minifies the source text (if needed)'''
    # there really isn't a good way to know if a file is already minified.
    # our heuristic is if source is more than 50 bytes greater of dest OR
    # if a hard return is found in the first 50 chars, we assume it is not minified.
    minified = minifier(text)
    if  abs(len(text) - len(minified)) > 50 or '\n' in text[:50]:
        return minified
    return text

